import logging
import requests
import json
import os

from queue import Queue
from threading import Thread
from time import time
from xml.etree.ElementTree import fromstring
from base64 import b64decode
from typing import TypedDict, List

from .types.track import ManifestMimeType


WORKERS_COUNT = 4

logger = logging.getLogger("download")


class Worker(Thread):
    def __init__(self, queue: Queue, function):
        Thread.__init__(self)
        self.queue = queue
        self.function = function
        self.daemon = True
        self.start()

    def run(self):
        while True:
            arg = self.queue.get()
            self.function(arg)
            self.queue.task_done()


class Threader:
    def __init__(self, workers_num: int, target, args: list) -> None:
        self.queue = Queue()

        for arg in args:
            self.queue.put(arg)

        self.workers: list[Worker] = [
            Worker(self.queue, target) for _ in range(workers_num)
        ]

    def run(self):
        ts = time()
        self.queue.join()
        return round(time() - ts, 2)


class Downloader:
    def __init__(self) -> None:
        self.indexed_content: list[tuple[int, bytes]] = []
        self.session = requests.Session()

    def download(self, urls: list[str]) -> bytes:
        indexed_urls = [(i, url) for (i, url) in enumerate(urls)]
        threader = Threader(WORKERS_COUNT, self._downloadFragment, indexed_urls)
        threader.run()
        sorted_content = sorted(self.indexed_content, key=lambda x: x[0])
        data = b"".join(content for _, content in sorted_content)
        return data

    def _downloadFragment(self, arg: tuple[int, str]):
        index, url = arg
        req = self.session.get(url)
        self.indexed_content.append((index, req.content))


def decodeManifest(manifest: str):
    return b64decode(manifest).decode()


def parseManifest(manifest: str):
    class AudioFileInfo(TypedDict):
        mimeType: str
        codecs: str
        encryptionType: str
        urls: List[str]

    data: AudioFileInfo = json.loads(manifest)
    return data


def parseManifestXML(xml_content: str):
    """
    Parses XML manifest file of the track.
    """

    NS = "{urn:mpeg:dash:schema:mpd:2011}"

    tree = fromstring(xml_content)

    representationElement = tree.find(
        f"{NS}Period/{NS}AdaptationSet/{NS}Representation"
    )
    if representationElement is None:
        raise ValueError("Representation element not found")

    codecs = representationElement.get("codecs")

    segmentElement = representationElement.find(f"{NS}SegmentTemplate")
    if segmentElement is None:
        raise ValueError("SegmentTemplate element not found")

    url_template = segmentElement.get("media")
    if url_template is None:
        raise ValueError("No `media` attribute in SegmentTemplate")

    timelineElements = segmentElement.findall(f"{NS}SegmentTimeline/{NS}S")
    if not timelineElements:
        raise ValueError("SegmentTimeline elements not found")

    total = 0
    for element in timelineElements:
        total += 1
        count = element.get("r")
        if count is not None:
            total += int(count)

    urls = [url_template.replace("$Number$", str(i)) for i in range(0, total + 1)]

    return urls, codecs


def showProgressBar(iteration: int, total: int, text: str, length=30):
    SQUARE, SQUARE_FILL = "□", "■"
    iteration_mb = iteration / 1024 / 1024
    total_mb = total / 1024 / 1024
    percent = 100 * (iteration / total)
    progress = int(length * iteration // total)
    bar = f"{SQUARE_FILL * progress}{SQUARE * (length - progress)}"
    print(
        f"\r{text} {bar} {percent:.0f}% {iteration_mb:.2f} / {total_mb:.2f} MB",
        end="\r",
    )
    if iteration >= total:
        print()


def download(url: str) -> bytes:
    logger.debug(url)
    # use session for performance
    session = requests.Session()
    req = session.get(url, stream=True)
    total_size = int(req.headers.get("content-length", 0))
    block_size = 1024 * 1024
    data = b""

    for block in req.iter_content(block_size):
        data += block
        showProgressBar(len(data), total_size, "Single URL")

    return data


def threadDownload(urls: list[str]) -> bytes:
    dl = Downloader()
    data = dl.download(urls)

    return data


def downloadTrackStream(
    full_path: str,
    encoded_manifest: str,
    mime_type: ManifestMimeType,
):
    logger.debug(f"mime_type: {mime_type}")
    manifest = decodeManifest(encoded_manifest)

    match mime_type:
        case "application/dash+xml":
            track_urls, codecs = parseManifestXML(manifest)
        case "application/vnd.tidal.bts":
            data = parseManifest(manifest)
            track_urls, codecs = data["urls"], data["codecs"]
        case _:
            raise ValueError(f"Unknown `mime_type`: {mime_type}")

    logger.debug(f"codecs: {codecs}")

    if len(track_urls) == 1:
        track_data = download(track_urls[0])
    else:
        track_data = threadDownload(track_urls)
        codecs = "mp4a"  # for tracks that have mulitple `mp4` fragments

    """
    known codecs
        flac (master)
        mp4a.40.2 (high)
        mp4a.40.5 (low)
    """

    if codecs is None:
        raise Exception("Missing codecs")

    if codecs == "flac":
        extension = "flac"
    elif codecs.startswith("mp4a"):
        extension = "m4a"
    else:
        extension = "flac"
        logger.warning(
            f'unknown file codecs: "{codecs}", please submit this as issue on GitHub'
        )

    file_path = os.path.dirname(full_path)
    file_name = f"{full_path}.{extension}"
    logger.debug(f"file_path: {file_path}, file_name: {file_name}")

    os.makedirs(file_path, exist_ok=True)

    with open(file_name, "wb+") as f:
        f.write(track_data)

    return file_name



import os
import time
import logging
from random import randint

from .api import TidalApi
from .auth import getDeviceAuth, getToken, refreshToken
from .config import Config
from .download import downloadTrackStream
from .parser import QUALITY_ARGS, parser
from .types import TRACK_QUALITY, TrackQuality, Track
from .utils import (
    RESOURCE,
    parseURL,
    formatFilename,
    loadingSymbol,
    setMetadata,
)


def main():
    logger = logging.getLogger("TIDDL")
    stream_handler = logging.StreamHandler()
    level_name_log = ""
    function_log = ""

    args = parser.parse_args()

    if args.silent:
        log_level = logging.WARNING
    elif args.verbose:
        log_level = logging.DEBUG
        level_name_log = "\033[1;34m%(levelname)s "
        function_log = " %(funcName)s"
    else:
        log_level = logging.INFO

    stream_handler.setLevel(log_level)

    if args.no_color:
        stream_handler.setFormatter(
            logging.Formatter(f"[ %(levelname)s ] %(name)s{function_log} - %(message)s")
        )
    else:
        stream_handler.setFormatter(
            logging.Formatter(
                f"{level_name_log}\033[1;95m%(name)s\033[0;35m{function_log}\033[0m %(message)s"
            )
        )

    config = Config()

    file_handler = logging.FileHandler(
        f"{config.config_directory}/tiddl.log", "a", "utf-8"
    )
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(
        logging.Formatter(
            "%(asctime)s %(levelname)s\t%(name)s.%(funcName)s %(message)s",
            datefmt="%x %X",
        )
    )

    logging.basicConfig(
        level=logging.DEBUG,
        handlers=[file_handler, stream_handler],
    )

    logger.debug(args)

    download_path = args.download_path or config["settings"]["download_path"]
    track_quality = (
        QUALITY_ARGS[args.quality]
        if args.quality
        else config["settings"]["track_quality"]
    )
    track_template = args.file_template or config["settings"]["track_template"]
    include_singles = args.include_singles

    if args.save_options:
        logger.info("saving new settings...")
        settings = config.update(
            {
                "settings": {
                    "download_path": download_path,
                    "track_quality": track_quality,
                    "track_template": track_template,
                }
            }
        ).get("settings")

        if settings:
            print("Current Settings:")
            for k, v in settings.items():
                print(f'> {k.upper()} "{v}"')

        logger.info(f"saved settings to {config.config_path}")

    if not config["token"]:
        auth = getDeviceAuth()
        text = f"> go to https://{auth['verificationUriComplete']} and add device!"
        expires_at = time.time() + auth["expiresIn"]
        i = 0
        while time.time() < expires_at:
            for _ in range(50):
                loadingSymbol(i, text)
                i += 1
                time.sleep(0.1)
            token = getToken(auth["deviceCode"])

            if token.get("error"):
                continue

            print()

            config.update(
                {
                    "token": token["access_token"],
                    "refresh_token": token["refresh_token"],
                    "token_expires_at": int(time.time()) + token["expires_in"],
                    "user": {
                        "user_id": str(token["user"]["userId"]),
                        "country_code": token["user"]["countryCode"],
                    },
                }
            )
            logger.info(f"authenticated!")
            break
        else:
            logger.info("time for authentication has expired")
            return

    t_now = int(time.time())
    token_expired = t_now > config["token_expires_at"]

    if token_expired:
        token = refreshToken(config["refresh_token"])
        config.update(
            {
                "token": token["access_token"],
                "token_expires_at": int(time.time()) + token["expires_in"],
            }
        )
        logger.info(f"refreshed token!")

    time_to_expire = config["token_expires_at"] - t_now
    days, hours = time_to_expire // (24 * 3600), time_to_expire % (24 * 3600) // 3600
    days_text = f" {days} {'day' if days == 1 else 'days'}" if days else ""
    hours_text = f" {hours} {'hour' if hours == 1 else 'hours'}" if hours else ""
    logger.debug(f"token expires in{days_text}{hours_text}")

    user_inputs: list[str] = args.input

    if len(user_inputs) == 0:
        logger.warning("no ID nor URL provided")
        return

    api = TidalApi(
        config["token"], config["user"]["user_id"], config["user"]["country_code"]
    )

    def downloadTrack(
        track: Track, file_template: str, skip_existing=True, sleep=False, playlist=""
    ):
        file_name = formatFilename(file_template, track, playlist)
        full_path = f"{download_path}/{file_name}"

        # it will stop detecting existing file for other extensions.
        # we need to store track `id + quality` in metadata to differentiate tracks
        # TODO: create better existing file detecting ✨
        if skip_existing and os.path.isfile(full_path + ".flac"):
            logger.info(f"already exists: {full_path}")
            return

        if sleep:
            sleep_time = randint(5, 15) / 10 + 1
            logger.info(f"sleeping for {sleep_time}s")
            time.sleep(sleep_time)

        stream = api.getTrackStream(track["id"], track_quality)
        quality = TRACK_QUALITY[stream["audioQuality"]]

        MASTER_QUALITIES: list[TrackQuality] = ["HI_RES_LOSSLESS", "LOSSLESS"]
        if stream["audioQuality"] in MASTER_QUALITIES:
            bit_depth, sample_rate = stream.get("bitDepth"), stream.get("sampleRate")
            if bit_depth is None or sample_rate is None:
                raise ValueError(
                    "bitDepth and sampleRate must be provided for master qualities"
                )
            details = f"{bit_depth} bit {sample_rate/1000:.1f} kHz"
        else:
            details = quality["details"]

        logger.info(f"{file_name} :: {quality['name']} Quality - {details}")

        track_path = downloadTrackStream(
            full_path,
            stream["manifest"],
            stream["manifestMimeType"],
        )

        logger.info(f"track saved in {track_path}")

        setMetadata(track_path, track)

    def downloadAlbum(album_id: str | int, skip_existing: bool):
        album = api.getAlbum(album_id)
        logger.info(f"album: {album['title']}")

        # i dont know if limit 100 is suspicious
        # but i will leave it here
        album_items = api.getAlbumItems(album_id, limit=100)
        for item in album_items["items"]:
            track = item["item"]
            downloadTrack(
                track,
                file_template=config["settings"]["album_template"],
                skip_existing=skip_existing,
                sleep=True,
            )

    skip_existing = not args.no_skip
    failed_input = []

    for user_input in user_inputs:
        input_type: RESOURCE
        input_id: str

        if user_input.isdigit():
            input_type = "track"
            input_id = user_input
        else:
            try:
                input_type, input_id = parseURL(user_input)
            except ValueError as e:
                logger.error(e)
                failed_input.append(user_input)
                continue

        match input_type:
            case "track":
                track = api.getTrack(input_id)
                downloadTrack(
                    track, file_template=track_template, skip_existing=skip_existing
                )
                continue

            case "album":
                downloadAlbum(input_id, skip_existing)
                continue

            case "artist":
                all_albums = []
                artist_albums = api.getArtistAlbums(input_id)
                all_albums.extend(artist_albums["items"])

                if include_singles:
                    artist_singles = api.getArtistAlbums(input_id, onlyNonAlbum=True)
                    all_albums.extend(artist_singles["items"])

                for album in all_albums:
                    downloadAlbum(album["id"], skip_existing)

                continue

            case "playlist":
                # TODO: add option to limit and set offset of playlist ✨
                playlist = api.getPlaylist(input_id)
                logger.info(f"playlist: {playlist['title']} ({playlist['url']})")

                playlist_items = api.getPlaylistItems(input_id)
                for item in playlist_items["items"]:
                    downloadTrack(
                        item["item"],
                        file_template=config["settings"]["playlist_template"],
                        skip_existing=skip_existing,
                        sleep=True,
                        playlist=playlist["title"],
                    )

                continue

            case _:
                logger.warning(f"invalid input: `{input_type}`")

        failed_input.append(input_id)

    if len(failed_input) > 0:
        logger.info(f"failed: {failed_input}")


if __name__ == "__main__":
    main()



import logging
from requests import Session

from .types import (
    SessionResponse,
    TrackQuality,
    Track,
    TrackStream,
    AristAlbumsItems,
    Album,
    AlbumItems,
    Playlist,
    PlaylistItems,
)

API_URL = "https://api.tidal.com/v1"


class TidalApi:
    def __init__(self, token: str, user_id: str, country_code: str) -> None:
        self.token = token
        self.user_id = user_id
        self.country_code = country_code

        self._session = Session()
        self._session.headers = {"authorization": f"Bearer {token}"}
        self._logger = logging.getLogger("TidalApi")

    def _request(self, endpoint: str, params={}):
        self._logger.debug(f"{endpoint} {params}")
        req = self._session.request(
            method="GET", url=f"{API_URL}/{endpoint}", params=params
        )

        # TODO: endpoints error handling ✨

        return req.json()

    def getSession(self) -> SessionResponse:
        return self._request(
            f"sessions",
        )

    def getTrackStream(self, id: str | int, quality: TrackQuality) -> TrackStream:
        return self._request(
            f"tracks/{id}/playbackinfo",
            {
                "audioquality": quality,
                "playbackmode": "STREAM",
                "assetpresentation": "FULL",
            },
        )

    def getTrack(self, id: str | int) -> Track:
        return self._request(f"tracks/{id}", {"countryCode": self.country_code})

    def getArtistAlbums(
        self, id: str | int, limit=50, offset=0, onlyNonAlbum=False
    ) -> AristAlbumsItems:
        params = {"countryCode": self.country_code, "limit": limit, "offset": offset}

        if onlyNonAlbum:
            params.update({"filter": "EPSANDSINGLES"})

        return self._request(
            f"artists/{id}/albums",
            params,
        )

    def getAlbum(self, id: str | int) -> Album:
        return self._request(f"albums/{id}", {"countryCode": self.country_code})

    def getAlbumItems(self, id: str | int, limit=10, offset=0) -> AlbumItems:
        return self._request(
            f"albums/{id}/items",
            {"countryCode": self.country_code, "limit": limit, "offset": offset},
        )

    def getPlaylist(self, uuid: str) -> Playlist:
        return self._request(
            f"playlists/{uuid}",
            {"countryCode": self.country_code},
        )

    def getPlaylistItems(self, uuid: str, limit=50, offset=0) -> PlaylistItems:
        return self._request(
            f"playlists/{uuid}/items",
            {"countryCode": self.country_code, "limit": limit, "offset": offset},
        )



import json
import logging

from pathlib import Path
from typing import TypedDict, Any
from .types import TrackQuality


class Settings(TypedDict, total=False):
    download_path: str
    track_quality: TrackQuality
    track_template: str
    album_template: str
    playlist_template: str


class User(TypedDict, total=False):
    user_id: str
    country_code: str


class ConfigData(TypedDict, total=False):
    token: str
    refresh_token: str
    token_expires_at: int
    settings: Settings
    user: User


CONFIG_FILENAME = ".tiddl_config.json"
DEFAULT_CONFIG: ConfigData = {
    "token": "",
    "refresh_token": "",
    "token_expires_at": 0,
    "settings": {
        "download_path": "tidal_download",
        "track_quality": "HIGH",
        "track_template": "{artist}/{title}",
        "album_template": "{artist}/{album}/{title}",
        "playlist_template": "{playlist}/{title}",
    },
    "user": {"user_id": "", "country_code": ""},
}


class Config:
    def __init__(self, config_path="") -> None:
        if config_path == "":
            home_directory = str(Path.home())
            self.config_directory = home_directory
        else:
            self.config_directory = config_path

        self.config_path = f"{self.config_directory}/{CONFIG_FILENAME}"
        self._config: ConfigData = DEFAULT_CONFIG
        self._logger = logging.getLogger("Config")

        try:
            with open(self.config_path, "r") as f:
                loaded_config: ConfigData = json.load(f)
                loaded_settings = loaded_config.get("settings")
                self._logger.debug(f"loaded {loaded_settings}")
                self.update(loaded_config)

        except FileNotFoundError:
            self._logger.debug("creating new file")
            self._save()  # save default config if file does not exist
            self._logger.debug("created new file")

    def _save(self) -> None:
        with open(self.config_path, "w") as f:
            self._logger.debug(self._config.get("settings"))
            json.dump(self._config, f, indent=2)

    def __getitem__(self, key: str) -> Any:
        return self._config[key]

    def __iter__(self):
        return iter(self._config)

    def __str__(self) -> str:
        return json.dumps(self._config, indent=2)

    def update(self, data: ConfigData) -> ConfigData:
        self._logger.debug("updating")
        merged_config: ConfigData = merge(data, self._config)
        self._config.update(merged_config)
        self._save()
        self._logger.debug("updated")
        return self._config.copy()


def merge(source, destination):
    # https://stackoverflow.com/a/20666342
    for key, value in source.items():
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            merge(value, node)
        else:
            destination[key] = value

    return destination



import os
import argparse

from .types import TRACK_QUALITY
from .types.track import TrackQuality


def shouldNotColor() -> bool:
    # TODO: add more checks ✨
    checks = ["NO_COLOR" in os.environ]
    return any(checks)


parser = argparse.ArgumentParser(
    description="\033[4mTIDDL\033[0m - Tidal Downloader",
    epilog="options defaults will be fetched from your config file.",
)

parser.add_argument(
    "input",
    type=str,
    nargs="*",
    help="track, album, playlist or artist - must be url, single id will be treated as track",
)

parser.add_argument(
    "-o",
    type=str,
    nargs="?",
    const=True,
    help="output file template, possible values are {id} {title} {number} {artist} {album} {artists}",
    dest="file_template",
)

parser.add_argument(
    "-p",
    type=str,
    nargs="?",
    const=True,
    help="download destination path",
    dest="download_path",
)

QUALITY_ARGS: dict[str, TrackQuality] = {
    details["arg"]: quality for quality, details in TRACK_QUALITY.items()
}

parser.add_argument(
    "-q",
    nargs="?",
    help="track quality",
    dest="quality",
    choices=QUALITY_ARGS.keys(),
)

parser.add_argument(
    "-is",
    help="include artist EPs and singles when downloading artist",
    dest="include_singles",
    action="store_true",
)

parser.add_argument(
    "-s",
    help="save options to config // show config file",
    dest="save_options",
    action="store_true",
)

parser.add_argument(
    "--no-skip",
    help="dont skip already downloaded tracks",
    action="store_true",
)

parser.add_argument(
    "--silent",
    help="silent mode",
    action="store_true",
)

parser.add_argument(
    "--verbose",
    help="show debug logs",
    action="store_true",
)

parser.add_argument(
    "--no-color",
    help="suppress output colors",
    action="store_true",
    default=shouldNotColor(),
)



import re
import os
import logging

from typing import TypedDict, Literal, List, get_args
from mutagen.flac import FLAC as MutagenFLAC
from mutagen.easymp4 import EasyMP4 as MutagenMP4

from .types.track import Track

RESOURCE = Literal["track", "album", "artist", "playlist"]
RESOURCE_LIST: List[RESOURCE] = list(get_args(RESOURCE))


logger = logging.getLogger("utils")


def parseURL(url: str) -> tuple[RESOURCE, str]:
    # remove trailing slash
    url = url.rstrip("/")
    # remove params
    url = url.split("?")[0]

    fragments = url.split("/")

    if len(fragments) < 2:
        raise ValueError(f"Invalid input: {url}")

    parsed_type, parsed_id = fragments[-2], fragments[-1]

    if parsed_type not in RESOURCE_LIST:
        raise ValueError(f"Invalid resource type: {parsed_type} ({url})")

    return parsed_type, parsed_id


class FormattedTrack(TypedDict):
    id: str
    title: str
    number: str
    artist: str
    album: str
    artists: str
    playlist: str


def formatFilename(template: str, track: Track, playlist="") -> str:
    artists = [artist["name"] for artist in track["artists"]]
    formatted_track: FormattedTrack = {
        "album": track["album"]["title"],
        "artist": track["artist"]["name"],
        "artists": ", ".join(artists),
        "id": str(track["id"]),
        "title": track["title"],
        "number": str(track["trackNumber"]),
        "playlist": playlist,
    }

    dirs = template.split("/")
    filename = dirs.pop().format(**formatted_track)

    template_without_filename = "/".join(dirs)
    formatted_dir = template_without_filename.format(**formatted_track)

    full_path = f"{sanitizeDirName(formatted_dir)}/{filename}"

    try:
        return full_path
    except KeyError as e:
        missing_key = e.args[0]
        raise ValueError(f"Missing key in track dictionary: {missing_key}")


def sanitizeDirName(dir_name):
    # replace invalid characters with an underscore
    return re.sub(r'[<>:"|?*]', "_", dir_name)


def loadingSymbol(i: int, text: str):
    symbols = ["⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏"]
    symbol = symbols[i % len(symbols)]
    print(f"\r{text} {symbol}", end="\r")


def setMetadata(file_path: str, track: Track):
    _, extension = os.path.splitext(file_path)

    if extension == ".flac":
        metadata = MutagenFLAC(file_path)
    elif extension == ".m4a":
        metadata = MutagenMP4(file_path)
    else:
        raise ValueError(f"Unknown file extension: {extension}")

    new_metadata: dict[str, str] = {
        "title": track["title"],
        "trackNumber": str(track["trackNumber"]),
        "discnumber": str(track["volumeNumber"]),
        "copyright": track["copyright"],
        "artist": track["artist"]["name"],
        "album": track["album"]["title"],
        "date": track["streamStartDate"][:10],
        # "tags": track["audioQuality"],
        # "id": str(track["id"]),
        # "url": track["url"],
    }

    try:
        metadata.update(new_metadata)
    except Exception as e:
        logger.error(e)
        return

    metadata.save()



import unittest

from .utils import parseURL, formatFilename
from .types.track import Track


class TestUtils(unittest.TestCase):

    def test_parseURL(self):
        self.assertEqual(
            parseURL("https://tidal.com/browse/track/284165609"), ("track", "284165609")
        )
        self.assertEqual(
            parseURL("https://tidal.com/browse/track/284165609/"),
            ("track", "284165609"),
        )
        self.assertEqual(
            parseURL("https://tidal.com/browse/track/284165609?u"),
            ("track", "284165609"),
        )
        self.assertEqual(
            parseURL(
                "https://listen.tidal.com/album/284165608/track/284165609",
            ),
            ("track", "284165609"),
        )

        self.assertEqual(
            parseURL("https://listen.tidal.com/album/284165608"), ("album", "284165608")
        )
        self.assertEqual(
            parseURL("https://tidal.com/browse/album/284165608"), ("album", "284165608")
        )
        self.assertEqual(
            parseURL("https://tidal.com/browse/album/284165608?u"),
            ("album", "284165608"),
        )

        self.assertEqual(
            parseURL("https://listen.tidal.com/artist/7695548"), ("artist", "7695548")
        )
        self.assertEqual(
            parseURL("https://tidal.com/browse/artist/7695548"), ("artist", "7695548")
        )

        self.assertEqual(
            parseURL(
                "https://tidal.com/browse/playlist/803be625-97e4-4cbb-88dd-43f0b1c61ed7"
            ),
            ("playlist", "803be625-97e4-4cbb-88dd-43f0b1c61ed7"),
        )
        self.assertEqual(
            parseURL(
                "https://listen.tidal.com/playlist/803be625-97e4-4cbb-88dd-43f0b1c61ed7"
            ),
            ("playlist", "803be625-97e4-4cbb-88dd-43f0b1c61ed7"),
        )

        self.assertEqual(
            parseURL(
                "https://listen.tidal.com/playlist/803be625-97e4-4cbb-88dd-43f0b1c61ed7"
            ),
            ("playlist", "803be625-97e4-4cbb-88dd-43f0b1c61ed7"),
        )

        self.assertEqual(parseURL("track/284165609"), ("track", "284165609"))
        self.assertEqual(
            parseURL("playlist/803be625-97e4-4cbb-88dd-43f0b1c61ed7"),
            ("playlist", "803be625-97e4-4cbb-88dd-43f0b1c61ed7"),
        )

        # we can also omit domain
        self.assertEqual(
            parseURL("playlist/803be625-97e4-4cbb-88dd-43f0b1c61ed7"),
            ("playlist", "803be625-97e4-4cbb-88dd-43f0b1c61ed7"),
        )

        self.assertRaises(ValueError, parseURL, "")

    def test_formatFilename(self):
        track: Track = {
            "id": 133017101,
            "title": "HAUTE COUTURE",
            "duration": 243,
            "replayGain": -7.7,
            "peak": 0.944031,
            "allowStreaming": True,
            "streamReady": True,
            "adSupportedStreamReady": True,
            "djReady": True,
            "stemReady": False,
            "streamStartDate": "2020-03-05T00:00:00.000+0000",
            "premiumStreamingOnly": False,
            "trackNumber": 1,
            "volumeNumber": 1,
            "version": None,
            "popularity": 29,
            "copyright": "2020 TUZZA Globale",
            "bpm": None,
            "url": "http://www.tidal.com/track/133017101",
            "isrc": "PL70D1900060",
            "editable": False,
            "explicit": False,
            "audioQuality": "LOSSLESS",
            "audioModes": ["STEREO"],
            "mediaMetadata": {"tags": ["LOSSLESS"]},
            "artist": {
                "id": 9550100,
                "name": "Tuzza Globale",
                "type": "MAIN",
                "picture": "125c9343-3257-407a-8285-5e9f1d283a2e",
            },
            "artists": [
                {
                    "id": 9550100,
                    "name": "Tuzza Globale",
                    "type": "MAIN",
                    "picture": "125c9343-3257-407a-8285-5e9f1d283a2e",
                },
                {
                    "id": 6847736,
                    "name": "Taco Hemingway",
                    "type": "FEATURED",
                    "picture": "7a1f5193-5d96-452c-b8dd-5ff0f81d5335",
                },
            ],
            "album": {
                "id": 133017100,
                "title": "HAUTE COUTURE",
                "cover": "efd381c2-a982-4d09-bb15-da872006cadf",
                "vibrantColor": "#f6a285",
                "videoCover": None,
            },
            "mixes": {"TRACK_MIX": "001ec78dae0d4a470999adefffd570"},
        }

        self.assertEqual(formatFilename("{title}", track), "HAUTE COUTURE")
        self.assertEqual(
            formatFilename("{artist} - {title}", track), "Tuzza Globale - HAUTE COUTURE"
        )
        self.assertEqual(
            formatFilename("{album} - {title}", track),
            "HAUTE COUTURE - HAUTE COUTURE",
        )
        self.assertEqual(
            formatFilename("{number}. {title}", track),
            "1. HAUTE COUTURE",
        )
        self.assertEqual(
            formatFilename("{artists} - {title}", track),
            "Tuzza Globale, Taco Hemingway - HAUTE COUTURE",
        )
        self.assertEqual(
            formatFilename("{id}", track),
            "133017101",
        )


if __name__ == "__main__":
    unittest.main()



from typing import TypedDict, Optional, List, Literal

from .track import Track


class ErrorResponse(TypedDict):
    status: int
    subStatus: int
    userMessage: str


class Client(TypedDict):
    id: int
    name: str
    authorizedForOffline: bool
    authorizedForOfflineDate: Optional[str]


class SessionResponse(TypedDict):
    sessionId: str
    userId: int
    countryCode: str
    channelId: int
    partnerId: int
    client: Client


class Items(TypedDict):
    limit: int
    offset: int
    totalNumberOfItems: int


class ArtistAlbum(TypedDict):
    id: int
    name: str
    type: Literal["MAIN"]


class Album(TypedDict):
    id: int
    title: str
    duration: int
    streamReady: bool
    streamStartDate: str
    allowStreaming: bool
    premiumStreamingOnly: bool
    numberOfTracks: int
    numberOfVideos: int
    numberOfVolumes: int
    releaseDate: str
    copyright: str
    type: str
    version: Optional[str]
    url: str
    cover: str
    videoCover: Optional[str]
    explicit: bool
    upc: str
    popularity: int
    audioQuality: str
    audioModes: List[str]
    artist: ArtistAlbum
    artists: List[ArtistAlbum]


class AristAlbumsItems(Items):
    items: List[Album]


class _AlbumTrack(TypedDict):
    item: Track
    type: Literal["track"]


class AlbumItems(Items):
    items: List[_AlbumTrack]


class _Creator(TypedDict):
    id: int


class Playlist(TypedDict):
    uuid: str
    title: str
    numberOfTracks: int
    numberOfVideos: int
    creator: _Creator
    description: str
    duration: int
    lastUpdated: str
    created: str
    type: str
    publicPlaylist: bool
    url: str
    image: str
    popularity: int
    squareImage: str
    promotedArtists: List[ArtistAlbum]
    lastItemAddedAt: str


class _PlaylistItem(TypedDict):
    item: Track
    type: Literal["track"]
    cut: Literal[None]


class PlaylistItems(Items):
    items: List[_PlaylistItem]



from typing import TypedDict, Literal

from .api import *
from .track import *

TrackArg = Literal["low", "normal", "high", "master"]


class QualityDetails(TypedDict):
    name: str
    details: str
    arg: TrackArg


TRACK_QUALITY: dict[TrackQuality, QualityDetails] = {
    "LOW": {"name": "Low", "details": "96 kbps", "arg": "low"},
    "HIGH": {"name": "Low", "details": "320 kbps", "arg": "normal"},
    "LOSSLESS": {"name": "High", "details": "16-bit, 44.1 kHz", "arg": "high"},
    "HI_RES_LOSSLESS": {
        "name": "Max",
        "details": "Up to 24-bit, 192 kHz",
        "arg": "master",
    },
}
